import numpy as np
import matplotlib.pyplot as plt
import pygame as pg
import os
import threading
from rlgw.model import WORLD, ENV
pg.init() # load pygame modules

#~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~#
# Game Module for Grid-World
# Implements graphics and most of pygame related subs 
#~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~ - ~#
        
class GAME:

    def __init__(self, env, print_scores=False):
        self.env = env
        self.R, self.C = self.env.R, self.env.C
        self.print_scores=print_scores
        
        # -for- pygame
        
        # size
        self.size_ratio = 50 # ratio of each cell size
        self.line_width = 2  # linewidth for gridlines 
        self.width = self.C*self.size_ratio   # Window width
        self.height = self.R*self.size_ratio  # Window height
        
        # images
        self_art_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)) , 'art')
        self.agent_img = pg.image.load(os.path.join(self_art_dir, 'agent.png'))
        self.posf_image = pg.image.load(os.path.join(self_art_dir, 'posf.png'))
        self.posi_image = pg.image.load( os.path.join(self_art_dir, 'posi.png'))
        
        # coordinate mapper
        ax, ay = self.agent_img.get_size()
        self.dx, self.dy = (self.size_ratio-ax)/2, (self.size_ratio-ay)/2
        self.mpos=128/self.env.world.RRP[1]
        self.mneg=128/self.env.world.RRN[0]
        
        # colors
        self.empty_col=(255,255,255)
        self.end_col=(255,255,0)
        self.linec=(0,0,0)
        self.bgcol=(0,50,100)
        self.hlast = {}
        self.reset_hlast()
        self.hrender = True
        self.font_family = 'consolas'
        self.font_size = int( 2.5*(self.C) )
        return
    def reset_hlast(self):
        self.hlast['tr'] = 0
        self.hlast['r'] = 0
        self.hlast['d'] = False
        self.hlast['ts'] = 0
        self.score_str = "[Score] | Steps : [0] | 0"
        return
    def update_hlast(self, total_rew, reward, done, steps):
        self.hlast['tr'] = total_rew
        self.hlast['r'] = reward
        self.hlast['d'] = done
        self.hlast['ts'] = steps
        self.score_str = "[Score] | Steps : ["+str(total_rew) + "] | "+str(steps)
        return
        
    def make_screen(self):
        """ Create game screen {OTC} """
        self.reset_hlast()
        # define font 
        # SysFont(name, size, bold=False, italic=False)
        self.font = pg.font.SysFont(self.font_family, min(self.font_size, 20 )) # after pygame is initialized
        
        # create a window or screen (pygame.display)
        self.screen = pg.display.set_mode(
            (self.width + self.line_width, self.height + self.line_width + self.size_ratio))  # screen display object
        self.screen.fill(self.bgcol)   # fill back ground color
        pg.display.set_caption(self.env.name)
        
        self.show_lines()              # create grid-lines ( one time call )
        self.show_grid()               # create grid-cells ( one time call )
        pg.display.update()
        return

    def show_lines(self):
        """ Draw the grid lines {OTC} """
        GRID = self.env.GRIDbox
        for i in range(self.R):
            for j in range(self.C):
                #  vertical lines
                pg.draw.line(self.screen, 
                                color=self.linec, 
                                start_pos=(j*self.size_ratio, 0), 
                                end_pos=(j*self.size_ratio, self.height), 
                             width=self.line_width)
            # end for (j)
            
            # last vertical line
            pg.draw.line(self.screen, 
                            color=self.linec, 
                            start_pos=((j+1)*self.size_ratio, 0), 
                            end_pos=((j+1)*self.size_ratio, 
                            self.height), 
                         width=self.line_width)
                         
            # horizontal lines
            pg.draw.line(self.screen, 
                            color=self.linec, 
                            start_pos=(0, i*self.size_ratio), 
                            end_pos=(self.width, i*self.size_ratio), 
                            width=self.line_width)
                            
        # end for (i)
        
        # last horizontal line
        pg.draw.line(self.screen, 
                        color=self.linec,
                        start_pos=(0, (i+1)*self.size_ratio), 
                        end_pos=(self.width, (i+1)*self.size_ratio), 
                        width=self.line_width)
        return
        
    def show_grid(self):
        """ Draw the grid cells {OTC} """
        GRID = self.env.GRIDbox
        for i in range(self.R):
            for j in range(self.C):
                if GRID[i,j]==self.env.world.ER:
                    colx= self.end_col
                elif GRID[i,j]==self.env.world.EMPTY:
                    colx= self.empty_col
                elif GRID[i,j]<0:
                    colx= (int(255-(GRID[i,j])*self.mneg),0,64)
                else:
                    colx= (0, int(255-(GRID[i,j])*self.mpos), 64)
                pg.draw.rect(self.screen, colx, pg.Rect( j*self.size_ratio+self.line_width, i*self.size_ratio+self.line_width, 
                                                          self.size_ratio-self.line_width*2, self.size_ratio-self.line_width*2))
        
       
    def make_update(self):
        self.update_grid()
        self.update_history()
        self.update_agent()
        self.update_score()
        pg.display.update()
        return
        
    def update_score(self):
    
        # redraw background
        pg.draw.rect(self.screen, self.bgcol, 
                     pg.Rect(self.line_width, 
                             self.R*self.size_ratio+self.line_width, 
                             self.width-self.line_width*2, 
                             self.size_ratio-self.line_width*2))
                             
        # display text using render(text, antialias, color, background=None)
        img_score = self.font.render(self.score_str, 1, self.empty_col, self.bgcol)
        self.screen.blit(img_score, (self.size_ratio/2, (self.R)*self.size_ratio+self.size_ratio/2))
        
        #img_reward = self.font.render("Steps: "+str(self.hlast['ts']), 1, self.empty_col, self.bgcol)
        #self.screen.blit(img_reward, (self.size_ratio*3+self.size_ratio/2, (self.R)*self.size_ratio+self.size_ratio/2))

    def update_agent(self):
        self.screen.blit(self.agent_img, (self.env.POS[1]*self.size_ratio+self.dy, self.env.POS[0]*self.size_ratio+self.dx))
        self.screen.blit(self.posi_image, (self.env.POSi[1]*self.size_ratio+self.dy, self.env.POSi[0]*self.size_ratio+self.dx))
        self.screen.blit(self.posf_image, (self.env.POSF[1]*self.size_ratio+self.dy, self.env.POSF[0]*self.size_ratio+self.dx))
         
    def update_grid(self):
        GRID = self.env.GRIDbox
        posL=   [(self.env.POS[0], self.env.POS[1]),
                (self.env.history[-1][0],self.env.history[-1][1]),] # only update the grid cells that changed - env.history will always have 1 item
        for i,j in posL:
            if GRID[i,j]==self.env.world.ER:
                colx= self.end_col
            elif GRID[i,j]==self.env.world.EMPTY:
                colx= self.empty_col
            elif GRID[i,j]<0:
                colx= (int(255-(GRID[i,j])*self.mneg),0, 64)
            else:
                colx= (0, int(255-(GRID[i,j])*self.mpos), 64)
            # redraw cells
            pg.draw.rect(self.screen, colx, pg.Rect( j*self.size_ratio+self.line_width, i*self.size_ratio+self.line_width, 
                                                      self.size_ratio-self.line_width*2, self.size_ratio-self.line_width*2))
        # draw the last move
        pg.draw.line(self.screen, color=self.end_col, 
                     start_pos=(posL[0][1]*self.size_ratio+self.size_ratio/2, posL[0][0]*self.size_ratio+self.size_ratio/2), 
                     end_pos=(posL[1][1]*self.size_ratio+self.size_ratio/2, posL[1][0]*self.size_ratio+self.size_ratio/2), 
                     width=1)
        
    def update_history(self):
        if self.hrender: # render history of moves (except for the most recent move that is displayed in update_grid())
            if len(self.env.history)>1:
                hist = np.array(self.env.history) * self.size_ratio + self.size_ratio/2
                hist[:,[0, 1]] = hist[:,[1, 0]]
                pg.draw.lines(self.screen, color=self.bgcol, closed=False, points=hist)


    def main(self):
        """ main game loop - run in a seprate thread/process and read the hlast dict for scores """
        # perpare
        self.env.reset()
        steps, total_rew, done = 0,0, self.env.in_done()
        self.make_screen() # will call pg.display.update()

        going = True
        while going:
            # ~ Check Events
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    going = False
                else:
                    if event.type == pg.KEYDOWN:
                        # any keypress to reset game in 'done' state
                        if event.key == pg.K_ESCAPE:
                            going =  False
                        else:
                            if done:
                                self.env.reset()
                                steps,total_rew, done = 0,0, self.env.in_done()
                                self.make_screen()
                            else:
                                if event.key == pg.K_KP5:
                                    a='XX'
                                elif event.key == pg.K_KP4:
                                    a='WW'
                                elif event.key == pg.K_KP6:
                                    a='EE'
                                elif event.key == pg.K_KP8:
                                    a='NN'
                                elif event.key == pg.K_KP2:
                                    a='SS'
                                elif event.key == pg.K_KP9:
                                    a='NE'
                                elif event.key == pg.K_KP7:
                                    a='NW'
                                elif event.key == pg.K_KP3:
                                    a='SE'
                                elif event.key == pg.K_KP1:
                                    a='SW'
                                else:
                                    a=""
                                    
                                if a in self.env.adict:
                                #if a>=0: # only on valid actions
                                    _, reward, done, _ = self.env.step(self.env.adict[a])
                                    total_rew+=reward
                                    steps+=1
                                    self.update_hlast(total_rew, reward, done, steps) # update scores
                                    if self.print_scores:
                                        if done:
                                            print(self.score_str)
            if going:
                self.make_update()   # ~ Render
            
            # ~ wait thread
            # pg.time.wait(1)
        
        # assert(going==False) #<-- test
        pg.display.quit()      # display quit here <-----------
        # pg.quit()            # unload pygame modules
        return




def play():
    """ Sample Game play in pygame window """
    game = GAME(
            env = ENV(
                    world = WORLD.get_world(
                            R=8, C=8, 
                            SP=(0,0), EP=(5,5), 
                            RNDSP=True, RNDEP=True, 
                            EMPTY=-1, RRN=[-8,-2], RRP=[2,8],  
                            RD=0.5, PRD=0.5, ER=0, 
                            MAXTS=8*8*2, 
                            SID=None, 
                            name="rlgw.default.world"
                                    ),
                    actions=ENV.ACTIONS,
                    penalty=(False, 0),  #<--- if True, imposes penalty on agent for staying in the same cell
                    always_reinit=True,
                    name="rlgw.default.env"  #<--- if True, always choose a new random initial state from distribution
                ), print_scores=True)
    
    
    
    th = threading.Thread(target=game.main)
    th.start()
    th.join()
    print('Done!')
    return