#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
import random, math
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from gym import spaces
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
class WORLD:
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    """ A Class for holding environment constants 
         Hint: use WORLD.get_world(...) to create WORLD objects """
         
    def get_world(R=4, C=4, SP=(0,0), EP=(3,3), RNDSP=False, RNDEP=False, EMPTY=-1, ER=0,
                  RRP=[5,10], RRN=[-10,-5], RD=0.5, PRD=0.5, MAXTS=4*4*2, SID = None, name=""):
        """
        Create a new WORLD object (which holds constants for creating environment instances)
        
        Args:
            R               Rows in the grid
            C               Cols in the grid
            SP              Start Position (relative to grid coordinate)
            EP              End Position (relative to grid coordinate)
            RNDSP           RaNDonmize Start Position (on call to env.reset())
            RNDEP           RaNDonmize End Position (on call to env.reset())
            EMPTY           Reward Value on Empty Cells
            RRP             Reward Range Positive (range of +ve rewards)
            RRN             Reward Range Negative (range of -ve rewards)
            RD              Reward Density (ratio of cells containing a reward v/s being empty)
            PRD             Positive Reward Density (ration of non-empty cells having a +ve reward value)
            ER              End Reward (special reward value on final position)
            MAXTS           Maximum time-steps that agent can take in a episode
            SID             Initial State Distribution (provide a list of initial states or keep None to grenerate from INIT_SV())
            name            World Name

        Returns:
            return          WORLD object
        """
    #~+~~+~~+~~+~~+~~+~~+~~~+~~+~~+~~+~~
        x = WORLD(name)
        x.R, x.C =          R, C
        x.SP, x.EP =        SP, EP
        x.RNDSP, x.RNDEP  = RNDSP, RNDEP
        x.EMPTY =           EMPTY
        x.RRP, x.RRN =      RRP, RRN
        x.RD =              RD
        x.PRD =             PRD
        x.ER =              ER
        x.MAXTS =           MAXTS
        x.SID =             SID
    #~+~~+~~+~~+~~+~~+~~+~~~+~~+~~+~~+~~
        return x  # WORLD instance
    #~+~~+~~+~~+~~+~~+~~+~~~+~~+~~+~~+~~
        
    # >-----------------------------------------------------------------------
    # > Other Required functions (__init__, empty_SV, done_SV, init_SV)
    # >----------------------------------------------------------------------- 
    def __init__(self, name="world"):
        self.name = name   
        # return __init__
        
    def empty_SV(self):
        """ Initializes a new empty state-vector - used by ENV.__init__() """
        return np.zeros(self.R * self.C + 4 , dtype='int')
        
    def done_SV(self, SV):
        """ Determines final states - used by ENV.in_done() """
        return not((SV[-2]!=SV[-4]) or (SV[-1]!=SV[-3]))
        
    def init_SV(self, SV):
        """ Generate a random initial state from a known Initial State Distribution - used by ENV._reinit() 
             >> returns: generates a new initial state if self.SID is not provided (is None) otherwise choose randomly from self.SID
             >> Note: argument SV is passed by ref, function returns nothing, it initializes SV with a choosen initial state """
        if type(self.SID)==type(None):
            #-------------------- [set start & end positions]
            spR, spC, epR, epC = 0, 0, 0, 0
            while (epR==spR and epC==spC):
                spR, spC = (random.randint(0, self.R-1), random.randint(0, self.C-1)) if (self.RNDSP) else (self.SP[0], self.SP[1])
                epR, epC = (random.randint(0, self.R-1), random.randint(0, self.C-1)) if (self.RNDEP) else (self.EP[0], self.EP[1])
            SV[-2], SV[-1], SV[-4], SV[-3]  = spR, spC, epR, epC  # copy on SV
            #-------------------- [set grid rewards]
            for i in range(len(SV)-4):
                    SV[i] = (random.randint(self.RRP[0],self.RRP[1]) \
                    if random.random() < self.PRD else random.randint(self.RRN[0],self.RRN[1])) \
                    if (random.random() < self.RD) else (self.EMPTY)
            #-------------------- [set special rewards]
            SV[spR*self.C + spC] = self.EMPTY
            SV[epR*self.C + epC] = self.ER
        else:
            np.copyto(SV, self.SID[random.randint(0, len(self.SID)-1)])
        return
    # >----------------------------------------------------------------------- 



# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ENV:
    """Simulates an environment based on underlying WORLD object
     must implement reset() and step(action) functions along with in_done()"""
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    ACTIONS = ('XX','NN', 'SS', 'EE', 'WW', 'NE', 'NW', 'SE', 'SW')
    # ------------------------------------------------------------------------
    # [1] Initialize - define actions (A), meta-vars and in_done()
    # ------------------------------------------------------------------------
    def __init__(self, world, actions, always_reinit=True, penalty=(False, 0), name="grid-world" ):
        
        if len(actions)<2:
            raise ValueError('At least 2 actions required!')
        for a in actions:
            if a not in ENV.ACTIONS:
                raise ValueError('Unknwon Action: '+str(a))
            
        # Perform some checks
        if world.R < 2 or world.C < 2 :
            raise ValueError('At least 2x2 grid required!')
        if (world.SP[0] == world.EP[0]) and (world.SP[1] == world.EP[1]):
            if not(world.RNDSP or world.RNDEP):
                raise ValueError('Start and End positions must be different!')
        if (world.RRP[0]>world.RRP[1]) or (world.RRN[0]>world.RRN[1]):
            raise ValueError('Reward Ranges are not ordered!')
        
        # initialize environment
        self.name = name
        self.world = world          # underlying WORLD object
        self.actions = actions
        #self.aexe = [self.XX, ]
        self.A = len(actions)               # no of actions (0,1...)
        self.adict = {}
        for i in range(self.A):
            self.adict[self.actions[i]]=i
        
        self.history=[]
        # ** state vectors **
        self.Si = self.world.empty_SV() # initial state vector
        self.S = self.world.empty_SV()  # current state vector
        self.done = True
        self.always_reinit=always_reinit
        self.penalty=penalty
        #-----------------
        self._built_META()
        #-----------------
    def _built_META(self):
        # copy variables required from self.world
        self.R, self.C = self.world.R, self.world.C
        self.EMPTY = self.world.EMPTY
        self.LEN = len(self.S)
        self.mapper_index=np.arange(0,self.LEN,1)
        self.MAXTS = self.world.MAXTS 
        self._max_episode_steps=self.world.MAXTS 
        self._elapsed_steps=0 
        
        
        self.observation_space = spaces.box.Box(-np.inf, np.inf, shape=(self.LEN,))
        self.action_space = spaces.discrete.Discrete(self.A)
        # other variables
        self.POSF = self.S[-4:-2] # r,c
        self.POS = self.S[-2:] # r,c
        self.GRID = self.S[0:-4] # grid
        self.POSi = self.Si[-2:] # r,c
        self.GRIDi = self.Si[0:-4] # grid
        self.GRIDbox = self.GRID.reshape(self.R,self.C)
        self.GRIDibox = self.GRIDi.reshape(self.R,self.C)
        # state initialization
        self._rr()
        self.clear_history = True
        # -for- PLOT
        self.FIGSR = 0.8
        self._x = np.arange(0.5,self.C+0.5)
        self._y = np.arange(0.5,self.R+0.5)
        self.vmin, self.vmax = self.world.RRN[0],self.world.RRP[1]
        self.cmap = 'PiYG'
        self.spcol = 'green'  # special color
        self.epcol = 'red'  # special color
        self.ccol = 'black' # current col
        self.acol = 'tab:grey' # arrow color
        
        return
    def get_state(self):
        return self.S[self.mapper_index]
    # ------------------------------------------------------------------------
    def clone(self, deep_copy=True):
        """ create a clone of current environemnt along with state vectors and underlying WORLD object
            if deep_copy==True then uses copy.deepcopy() to create a copy of underlying WORLD object instead of passing reference"""
        my_world = None
        if deep_copy:
            my_world=deepcopy(self.world)
        else:
            my_world=self.world
            
        res=ENV(world=my_world, 
                name=self.name, 
                always_reinit=self.always_reinit, 
                penalty=self.penalty, actions=self.actions)
        np.copyto(res.S, self.S)
        np.copyto(res.Si, self.Si)
        res.done = self.done
        res._elapsed_steps=self._elapsed_steps
        return res
        
    # ------------------------------------------------------------------------
    # [2] State Space - define in_done(), reinit() and reset()
    # ------------------------------------------------------------------------
    
    def in_done(self):
        """ detemines if env is in terminal state """
        return self.world.done_SV(self.S) # link the done_SV functions from world object
        
    def reset(self):
        """Resets: copy initial state Si to current state S
        If self.always_reinit=True then randomize Initial state Si before copying to S
            otherwise copies the last Initial state generated by self,_reinit()"""
        _ = self._reinit() if self.always_reinit else None
        return self._reset()

    def _reinit(self):
        """ helper: choose a new initial state from initial state distribution """
        self.world.init_SV(self.Si)
        return self.Si
        
    def _reset(self):
        """ helper: copy last choosen initial state to current state vector i.e. reset to last known initial state """
        np.copyto(self.S, self.Si)
        self.done = self.in_done()
        self.history.clear()
        self.history.append([self.POS[0], self.POS[1]])
        self._elapsed_steps=0
        return self.get_state()
        
    def _rr(self):
        _ = self._reinit()
        _ = self._reset()
        return
        
    def random_action(self):
        return random.randint(0, self.A-1)
        
    # ------------------------------------------------------------------------
    # [3] Step - define step(action) function - returns reward, done
    # ------------------------------------------------------------------------
    
    def XX(self):
        pass
    def NN(self):
        if self.POS[0] != 0: 
            self.POS[0] = self.POS[0] - 1
    def EE(self):
        if self.POS[1] != self.C-1: 
            self.POS[1] = self.POS[1] + 1
    def WW(self):
        if self.POS[1] != 0: 
            self.POS[1] = self.POS[1] - 1
    def SS(self):
        if self.POS[0] != self.R-1: 
            self.POS[0] = self.POS[0] + 1
    def NE(self):
        if self.POS[0] != 0 and self.POS[1] != self.C-1: 
            self.POS[0] = self.POS[0] - 1
            self.POS[1] = self.POS[1] + 1
    def NW(self):
        if self.POS[0] != 0 and self.POS[1] != 0: 
            self.POS[0] = self.POS[0] - 1
            self.POS[1] = self.POS[1] - 1
    def SE(self):
        if self.POS[0] != self.R-1 and self.POS[1] != self.C-1: 
            self.POS[0] = self.POS[0] + 1
            self.POS[1] = self.POS[1] + 1
    def SW(self):
        if self.POS[0] != self.R-1 and self.POS[1] !=0: 
            self.POS[0] = self.POS[0] + 1
            self.POS[1] = self.POS[1] - 1
                
    def step(self, action):
        if self.done:
            raise ValueError('Cannot step in terminal state!')
        #----------------------------~~~~~~~@
        #----------------------------~~~~~~~>[ meta info (BEFORE state transition) ]
        if self.clear_history:
            self.history.clear()
            self.clear_history=False
        self.history.append( [self.POS[0],self.POS[1]] )
        #----------------------------~~~~~~~!
        #----------------------------~~~~~~~>>[ state transition begin ]
        act = self.actions[action]
        getattr(self, act)()
        penalty=0
        if self.penalty[0]:
            px,py = self.history[-1][0], self.history[-1][1]
            cx,cy = self.POS[0], self.POS[1]
            if px-cx==0 and py-cy==0:
                penalty=self.EMPTY*self.penalty[1]
            else:
                penalty=0

        ixd = self.POS[0]*self.C + self.POS[1]
        reward = self.GRID[ixd]     #<<---- collect reward
        self.GRID[ixd] = self.EMPTY #<<---- change state vector
        self.done = self.in_done()  #<<---- update self.done
        #----------------------------~~~~~~~>>[ state transition complete ]
        #----------------------------~~~~~~~>[ meta info (AFTER state transition) ]
        self.clear_history=self.done
        #----------------------------~~~~~~~>[ define any info here ]
        #info = penalty
        #----------------------------~~~~~~~@
        self._elapsed_steps+=1
        return self.get_state(), reward+penalty, self.done, penalty
        #----------------------------~~~~~~~@
        
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

    def tostr(self):
        """ Text based render """
        rr,cc = self.POS[0], self.POS[1]
        res=""
        for r in range(self.R):
            for c in range(self.C):
                jes = str(self.GRID[r*self.C+c])
                if r==rr and c==cc:
                    jes = "["+jes+"]\t"
                else:
                    jes = jes+"\t"
                res+=jes
            res+="\n"
        print(res)
        return

    def render(self, mode=1):
        """ matplotlib based render """
        if mode<1:
            self.tostr()
            return
    
        fig_ratio=self.FIGSR
        fig, ax = plt.subplots(1,2 , figsize=(2*fig_ratio*self.C ,fig_ratio*self.R ))
        for i in range(0,2):
            ax[i].clear()
            ax[i].set_xticks(self._x)
            ax[i].set_yticks(self._y)
            ax[i].set_xticklabels([])
            ax[i].set_yticklabels([])
            ax[i].grid(axis='both')
            
        ax[0].imshow(self.GRIDbox, cmap=self.cmap, vmin=self.vmin,vmax=self.vmax)
        ax[1].imshow(self.GRIDibox, cmap=self.cmap, vmin=self.vmin,vmax=self.vmax)        
        for (j,i),label in np.ndenumerate(self.GRIDibox):
            ax[1].text(i,j,label,ha='left',va='top')

        # Current, Initial and Final POS
        ax[0].scatter(self.POS[1], self.POS[0], color=self.ccol,marker='o')
        ax[1].scatter(self.POS[1], self.POS[0], color=self.ccol,marker='o')
        ax[0].scatter(self.POSi[1], self.POSi[0], color=self.spcol,marker='.')
        ax[1].scatter(self.POSi[1], self.POSi[0], color=self.spcol,marker='.')
        ax[0].scatter(self.POSF[1], self.POSF[0], color=self.epcol, marker='.')
        ax[1].scatter(self.POSF[1], self.POSF[0], color=self.epcol, marker='.')
        if len(self.history)>0:
            sh = self.history[-1]
            ax[0].annotate("", 
                           xy=( self.POS[1], self.POS[0] ),
                           xytext=(sh[1],sh[0]), 
                  arrowprops=dict(arrowstyle="->", linewidth=0.8, color=self.ccol))
            ax[1].annotate("", 
                           xy=( self.POS[1], self.POS[0] ),
                           xytext=(sh[1],sh[0]), 
                  arrowprops=dict(arrowstyle="->", linewidth=0.8, color=self.ccol))
            #if self.done:
            for k in range(1, len(self.history)):
                si, sj = self.history[k-1],self.history[k] 
                ax[0].annotate("", xy=(sj[1], sj[0]), xytext=(si[1], si[0]), 
                                arrowprops=dict(arrowstyle="->", linewidth=0.5, color=self.acol))
                ax[1].annotate("", xy=(sj[1], sj[0]), xytext=(si[1], si[0]), 
                                arrowprops=dict(arrowstyle="->", linewidth=0.5, color=self.acol))
        plt.show()
        return fig

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
     
    def get_env_batch(batch_size, base_env, render_envs=False):
        """ returns a batch of environments for testing purposes """
        batch=[]
        for i in range(batch_size):
            _env = base_env.clone(deep_copy=False)
            _env._rr()
            batch.append(_env)
        return batch
    
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-